@@ -269,14 +269,25 @@ and subtle differences that must be taken into consideration:
     udp_unicast_join_hook/3,                    % +Scope, +From, +Data
     black_list/1.                               % +Term
 
-:- meta_predicate safely(0).
+:- meta_predicate
+    safely(0),
+    safely_det(0).
 
 safely(Predicate) :-
+    Err = error(_,_),
     catch(Predicate, Err,
-          (   Err == '$aborted'
-          ->  !, fail
-          ;   print_message(error, Err), fail
-          )).
+          print_message_fail(Err)).
+
+safely_det(Predicate) :-
+    Err = error(_,_),
+    catch(Predicate, Err,
+          print_message_fail(Err)),
+    !.
+safely_det(_).
+
+print_message_fail(Term) :-
+    print_message(error, Term),
+    fail.
 
 udp_broadcast_address(IPAddress, Subnet, BroadcastAddress) :-
     IPAddress = ip(A1, A2, A3, A4),
@@ -495,7 +506,7 @@ ld_dispatch(S, request(Key, Term), From, Scope) :-
            safely((udp_term_string(Scope, reply(Key,Term), Message),
                    udp_send(S, Message, From, [])))).
 ld_dispatch(_S, send(Term), _From, _Scope) :-
-    safely(broadcast(Term)).
+    safely_det(broadcast(Term)).
 ld_dispatch(_S, reply(Key, Term), From, _Scope) :-
     (   reply_queue(Key, Queue)
     ->  safely(thread_send_message(Queue, Term:From))
@@ -806,6 +817,32 @@ peer_address(IP, Scope, IPAddress:Port) :-
 		 *             HOOKS		*
 		 *******************************/
 
+%!  udp_term_string_hook(+Scope, +Term, -String) is det.
+%!  udp_term_string_hook(+Scope, -Term, +String) is semidet.
+%
+%   Hook  for  serializing  the  message    Term.   The  default  writes
+%   =|%prolog\n|=, followed by the Prolog term  in quoted notation while
+%   ignoring operators. This hook may use alternative serialization such
+%   as fast_term_serialized/2, use  library(ssl)   to  realise encrypted
+%   messages, etc.
+%
+%   @arg Scope is the scope for which the message is broadcasted.  This
+%   can be used to use different serialization for different scopes.
+%   @arg Term encapsulates the term broadcasted by the application as
+%   follows:
+%
+%     - send(ApplTerm)
+%       Is sent by broadcast(udp(Scope, ApplTerm))
+%     - request(Id,ApplTerm)
+%       Is sent by broadcast_request/1, where Id is a unique large
+%       (64 bit) integer.
+%     - reply(Id,ApplTerm)
+%       Is sent to reply on a broadcast_request/1 request that has
+%       been received.  Arguments are the same as above.
+%
+%   @throws The hook may throw udp(invalid_message) to stop processing
+%   the message.
+
 %!  udp_term_string(+Scope, +Term, -String) is det.
 %!  udp_term_string(+Scope, -Term, +String) is semidet.
 %
@@ -816,17 +853,17 @@ peer_address(IP, Scope, IPAddress:Port) :-
 %   In mode (+,-), Term is written with the options ignore_ops(true) and
 %   quoted(true).
 %
-%   This predicate first calls  udp_term_string_hook/3   with  the  same
-%   signature. This hook  may  use   alternative  serialization  such as
-%   fast_term_serialized/2,  use  library(ssl)  to    realise  encrypted
-%   messages, etc.
-%
-%   @arg Scope is the scope for which the message is broadcasted.  This
-%   can be used to use different serialization for different scopes.
+%   This predicate first calls  udp_term_string_hook/3.
 
 udp_term_string(Scope, Term, String) :-
-    udp_term_string_hook(Scope, Term, String),
-    !.
+    catch(udp_term_string_hook(Scope, Term, String), udp(Error), true),
+    !,
+    (   var(Error)
+    ->  true
+    ;   Error == invalid_message
+    ->  fail
+    ;   throw(udp(Error))
+    ).
 udp_term_string(_Scope, Term, String) :-
     (   var(String)
     ->  format(string(String), '%-prolog-\n~W',
