@@ -3,7 +3,7 @@
     Author:        Jeffrey Rosenwald, Jan Wielemaker
     E-mail:        jeffrose@acm.org
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2009-2018, Jeffrey Rosenwald
+    Copyright (c)  2009-2019, Jeffrey Rosenwald
                    CWI, Amsterdam
     All rights reserved.
 
@@ -44,6 +44,9 @@
             paxos_on_change/3,                  % ?Key, ?Value, +Goal
 
             paxos_initialize/1,			% +Options
+
+            paxos_property/1,                   % ?Property
+            paxos_quorum_ask/4,                 % ?Templ, +Msg, -Result, +Options
                                                 % Hook support
             paxos_replicate_key/3               % +Nodes, ?Key, +Options
           ]).
@@ -53,6 +56,7 @@
 :- use_module(library(settings)).
 :- use_module(library(option)).
 :- use_module(library(error)).
+:- use_module(library(apply)).
 :- use_module(library(solution_sequences)).
 
 /** <module> A Replicated Data Store
@@ -130,7 +134,8 @@ identical for all attentive members of the quorum.|_
     paxos_on_change(?, ?, 0).
 
 :- multifile
-    paxos_message_hook/3.               % +PaxOS, +TimeOut, -Message
+    paxos_message_hook/3,               % +PaxOS, +TimeOut, -Message
+    paxos_ledger_hook/5.                % +Op, ?Key, ?Gen, ?Value, ?Status
 
 :- setting(max_sets, nonneg, 20,
            "Max Retries to get to an agreement").
@@ -143,7 +148,8 @@ identical for all attentive members of the quorum.|_
 :- setting(death_half_life, number, 10,
            "Half-time for failure score").
 :- setting(death_score, number, 100,
-           "Number of keys replicated per second").
+           "Consider a node dead if cummulative failure \c
+            score exceeds this number").
 
 
 %!  paxos_initialize(+Options) is det.
@@ -165,34 +171,28 @@ identical for all attentive members of the quorum.|_
 %     NodeID must be a small non-negative integer as these identifiers
 %     are used in bitmaps.
 
-paxos_initialize(Options) :-
-    paxos_assign_node(Options),
-    paxos_initialize.
-
-%!  paxos_initialize is det.
-%
-%   causes any required runtime initialization to occur. It is called as
-%   a side-effect of initialize/0, which is now required as part of
-%   an applications initialization directive.
-
 :- dynamic  paxos_initialized/0.
 :- volatile paxos_initialized/0.
 
-paxos_initialize :-
+paxos_initialize(_Options) :-
     paxos_initialized,
     !.
-paxos_initialize :-
-    with_mutex(paxos, paxos_initialize_sync).
+paxos_initialize(Options) :-
+    with_mutex(paxos, paxos_initialize_sync(Options)).
 
-paxos_initialize_sync :-
+paxos_initialize_sync(_Options) :-
     paxos_initialized,
     !.
-paxos_initialize_sync :-
+paxos_initialize_sync(Options) :-
+    at_halt(paxos_leave),
     listen(paxos, paxos(X), paxos_message(X)),
-    paxos_assign_node([]),
+    paxos_assign_node(Options),
     start_replicator,
     asserta(paxos_initialized).
 
+paxos_initialize :-
+    paxos_initialize([]).
+
 
 		 /*******************************
 		 *            ADMIN		*
@@ -215,6 +215,11 @@ paxos_set_admin(Name, Value) :-
     admin_key(Name, Key),
     paxos_set(Key, Value).
 
+paxos_set_admin_bg(Name, Value) :-
+    thread_create(ignore(paxos_set_admin(Name, Value)), _,
+                  [ detached(true)
+                  ]).
+
 
 		 /*******************************
 		 *           NODE DATA		*
@@ -243,58 +248,97 @@ paxos_set_admin(Name, Value) :-
     quorum/1,                           % Bitmap
     failed/1,                           % Bitmap
     failed/3,                           % NodeID, LastTried, Score
-    dead/1.                             % Bitmap
+    leaving/0,                          % Node is leaving
+    dead/1,                             % Bitmap
+    salt/1.                             % Unique key
 :- volatile
     node/1,
     quorum/1,
-    dead/1,
     failed/1,
-    failed/3.
+    failed/3,
+    leaving/0,
+    dead/1,
+    salt/1.
 
 %!  paxos_assign_node(+Options) is det.
 %
 %   Assign a node for this  paxos  instance.   If  node  is  given as an
 %   option, this is the node id that   is used. Otherwise the network is
-%   analyses and the system selects a new node.
-%
-%   @bug The current implementation does not   avoid that two nodes that
-%   are added at the same time get the same identifier.
+%   analysed and the system selects a new node.
 
-paxos_assign_node([]) :-
-    node(_),                            % already done
+paxos_assign_node(Options) :-
+    (   option(node(Node), Options)
+    ->  node(Node)
+    ;   node(_)
+    ),                                          % already done
     !.
 paxos_assign_node(Options) :-
+    between(1, 20, Retry),
     option(node(Node), Options, Node),
     (   node(_)
     ->  permission_error(set, paxos_node, Node)
     ;   true
     ),
-    debug(paxos(node), 'Finding peers ...', []),
-    paxos_message(node(N,Q,D), 0.25, NodeQuery),
-    findall(t(N,Q,D),
+    retractall(dead(_)),
+    retractall(quorum(_)),
+    retractall(failed(_)),
+    retractall(failed(_,_,_)),
+    retractall(leaving),
+    Salt is random(1<<63),
+    asserta(salt(Salt)),
+    paxos_message(node(N,Q,D):From, 0.25, NodeQuery),
+    findall(t(N,Q,D,From),
             broadcast_request(NodeQuery),
-            NodeStatus),
-    arg_union(2, NodeStatus, Quorum),
-    arg_union(3, NodeStatus, Dead),
+            Network),
+    select(t(self,0,Salt,Me), Network, AllNodeStatus),
+    partition(starting, AllNodeStatus, Starting, Running),
+    nth_starting(Starting, Salt, Offset),
+    retractall(salt(_)),
+    debug(paxos(node), 'Me@~p; starting: ~p; running: ~p',
+          [Me, Starting, Running]),
+    arg_union(2, Running, Quorum),
+    arg_union(3, Running, Dead),
     (   var(Node)
-    ->  (   between(0, 1000, Node),
-            \+ memberchk(t(Node,_,_), NodeStatus),
-            Dead /\ (1<<Node) =:= 0
+    ->  (   call_nth(( between(0, 1000, Node),
+                       \+ memberchk(t(Node,_,_,_), Running),
+                       Dead /\ (1<<Node) =:= 0),
+                     Offset)
         ->  debug(paxos(node), 'Assigning myself node ~d', [Node])
         ;   resource_error(paxos_nodes)
         )
-    ;   memberchk(t(Node,_,_), NodeStatus)
+    ;   memberchk(t(Node,_,_,_), Running)
     ->  permission_error(set, paxos_node, Node)
     ;   Rejoin = true
     ),
     asserta(node(Node)),
-    asserta(dead(Dead)),
-    set_quorum(Node, Quorum),
-    (   Rejoin == true
-    ->  paxos_rejoin
-    ;   true
+    (   claim_node(Node, Me)
+    ->  !,
+        asserta(dead(Dead)),
+        set_quorum(Node, Quorum),
+        (   Rejoin == true
+        ->  paxos_rejoin
+        ;   true
+        )
+    ;   debug(paxos(node), 'Node already claimed; retrying (~p)', [Node, Retry]),
+        retractall(node(Node)),
+        fail
     ).
 
+starting(t(self,_Quorum,_Salt,_Address)).
+
+nth_starting(Starting, Salt, N) :-
+    maplist(arg(3), Starting, Salts),
+    sort([Salt|Salts], Sorted),
+    nth1(N, Sorted, Salt),
+    !.
+
+claim_node(Node, Me) :-
+    paxos_message(claim_node(Node, Ok):From, 0.25, NodeQuery),
+    forall((   broadcast_request(NodeQuery),
+               From \== Me,
+               debug(paxos(node), 'Claim ~p ~p: ~p', [Node, From, Ok])
+           ),
+           Ok == true).
 
 set_quorum(Node, Quorum0) :-
     Quorum is Quorum0 \/ (1<<Node),
@@ -313,11 +357,13 @@ set_quorum(Node, Quorum0) :-
 paxos_rejoin :-
     node(Node),
     repeat,
-        paxos_get_admin(dead, Dead0),
-        Dead is Dead0 /\ \(1<<Node),
-        (   Dead == Dead0
-        ->  true
-        ;   paxos_set_admin(dead, Dead)
+        (   paxos_get_admin(dead, Dead0)
+        ->  Dead is Dead0 /\ \(1<<Node),
+            (   Dead == Dead0
+            ->  true
+            ;   paxos_set_admin(dead, Dead)
+            )
+        ;   true
         ),
     !.
 
@@ -329,14 +375,17 @@ paxos_rejoin :-
 %   paxos_leave/1 version is called  to  discard   other  nodes  if they
 %   repeatedly did not respond to queries.
 
-:- at_halt(paxos_leave).
-
 paxos_leave :-
     node(Node),
+    !,
+    asserta(leaving),
     paxos_leave(Node),
     Set is 1<<Node,
     paxos_message(forget(Set), -, Forget),
-    broadcast(Forget).
+    broadcast(Forget),
+    unlisten(paxos),
+    retractall(leaving).
+paxos_leave.
 
 paxos_leave(Node) :-
     !,
@@ -356,6 +405,8 @@ paxos_update_set(Set, How) :-
       (   Value == NewValue
       ->  true
       ;   paxos_set_admin(Set, NewValue)
+      ->  true
+      ;   leaving
       ),
     !.
 
@@ -450,32 +501,74 @@ life_quorum(Quorum, LifeQuorum) :-
 
 :- admin_key(quorum, Key),
    listen(paxos_changed(Key, Quorum),
-          update_copy(quorum, Quorum)).
+          update_quorum(Quorum)).
 :- admin_key(dead, Key),
    listen(paxos_changed(Key, Death),
-          update_copy(death, Death)).
+          update_dead(Death)).
 
-update_copy(Name, Value) :-
-    Term =.. [Name,Value0],
-    clause(Term, true, Ref),
+update_quorum(Proposed) :-
+    debug(paxos(node), 'Received quorum proposal 0x~16r', [Proposed]),
+    quorum(Proposed),
+    !.
+update_quorum(Proposed) :-
+    leaving,
     !,
-    (   Value0 == Value
-    ->  true
-    ;   debug(paxos(node), 'Updated ~w to 0x~16r', [Name, Value]),
-        NewTerm =.. [Name,Value],
-        updated(NewTerm, Value0),
-        asserta(NewTerm),
+    update(quorum(Proposed)).
+update_quorum(Proposed) :-
+    node(Node),
+    Proposed /\ (1<<Node) =\= 0,
+    !,
+    update(quorum(Proposed)).
+update_quorum(Proposed) :-
+    node(Node),
+    NewQuorum is Proposed \/ (1<<Node),
+    update(quorum(NewQuorum)),
+    debug(paxos(node), 'I''m not in the quorum! Proposing 0x~16r', [NewQuorum]),
+    paxos_set_admin_bg(quorum, NewQuorum).
+
+update_dead(Proposed) :-
+    debug(paxos(node), 'Received dead proposal 0x~16r', [Proposed]),
+    dead(Proposed),
+    !.
+update_dead(Proposed) :-
+    leaving,
+    !,
+    update(dead(Proposed)).
+update_dead(Proposed) :-
+    node(Node),
+    Proposed /\ (1<<Node) =:= 0,
+    !,
+    update(dead(Proposed)).
+update_dead(Proposed) :-
+    node(Node),
+    NewDead is Proposed /\ \(1<<Node),
+    update(dead(NewDead)),
+    paxos_set_admin_bg(dead, NewDead).
+
+update(Clause) :-
+    functor(Clause, Name, Arity),
+    functor(Generic, Name, Arity),
+    (   clause(Generic, true, Ref)
+    ->  asserta(Clause),
         erase(Ref)
+    ;   asserta(Clause)
     ).
-update_copy(Name, Value) :-
-    NewTerm =.. [Name,Value],
-    asserta(NewTerm).
 
-updated(quorum(Quorum), Quorum0) :-
-    Quorum /\ \Quorum0 =\= 0,
-    !,
-    start_replicator.
-updated(_, _).
+%!  paxos_property(?Property)
+%
+%   True if Property is  a  current   property  for  the  paxos network.
+%   Defined properties are:
+%
+%     - node(?NodeID)
+%     - quorum(?NodeBitmap)
+%     - failed(?NodeBitmap)
+
+paxos_property(node(NodeID)) :-
+    node(NodeID).
+paxos_property(quorum(Quorum)) :-
+    quorum(Quorum).
+paxos_property(failed(Nodes)) :-
+    failed(Nodes).
 
 
 		 /*******************************
@@ -508,8 +601,10 @@ updated(_, _).
 %     our node id and the generation.
 %     - forget(+Nodes)
 %     Forget the existence of Nodes.
-%     - node(-Node)
-%     Get the node id.
+%     - node(-Node,-Quorum,-Dead)
+%     Get my view about the network.  Node is the (integer) node id of
+%     this node, Quorum is the idea of the quorum and Dead is the idea
+%     about non-responsive nodes.
 %
 %   @tbd: originally the changed was  handled  by   a  get  and when not
 %   successful with a new set, named   _paxos_audit_. I don't really see
@@ -525,14 +620,14 @@ paxos_message(prepare(Key,Node,Gen,Value)) :-
     debug(paxos, 'Prepared ~p-~p@~d', [Key,Value,Gen]).
 paxos_message(accept(Key,Node,Gen,GenA,Value)) :-
     node(Node),
-    debug(paxos, 'Accept ~p-~p@~p?', [Key, Value, Gen]),
     (   ledger_update(Key, Gen, Value)
     ->  debug(paxos, 'Accepted ~p-~p@~d', [Key,Value,Gen]),
         GenA = Gen
-    ;   debug(paxos, 'Rejected ~p@~d', [Key, Gen]),
+    ;   debug(paxos, 'Rejected ~p-~p@~d', [Key,Value,Gen]),
         GenA = nack
     ).
 paxos_message(changed(Key,Gen,Value,Acceptors)) :-
+    debug(paxos, 'Changed ~p-~p@~d for ~p', [Key, Value, Gen, Acceptors]),
     ledger_update_holders(Key,Gen,Acceptors),
     broadcast(paxos_changed(Key,Value)).
 paxos_message(learn(Key,Node,Gen,GenA,Value)) :-
@@ -555,9 +650,23 @@ paxos_message(retrieve(Key,Node,K,Value)) :-
 paxos_message(forget(Nodes)) :-
     ledger_forget(Nodes).
 paxos_message(node(Node,Quorum,Dead)) :-
+    (   node(Node),
+        quorum(Quorum),
+        dead(Dead)
+    ->  true
+    ;   salt(Salt),
+        Node = self,
+        Quorum = 0,
+        Dead = Salt
+    ).
+paxos_message(claim_node(Node, Ok)) :-
+    (   node(Node)
+    ->  Ok = false
+    ;   Ok = true
+    ).
+paxos_message(ask(Node, Message)) :-
     node(Node),
-    quorum(Quorum),
-    dead(Dead).
+    broadcast_request(Message).
 
 
 		 /*******************************
@@ -617,9 +726,12 @@ paxos_set(Key, Value, Options) :-
     paxos_message(prepare(Key,Np,Rp,Value), TMO, Prepare),
     between(0, Retries, _),
       life_quorum(Quorum, Alive),
+      Alive \== 0,
+      debug(paxos, 'Set: ~p -> ~p', [Key, Value]),
       collect(Quorum, false, Np, Rp, Prepare, Rps, PrepNodes),
+      debug(paxos, 'Set: quorum: 0x~16r, prepared by 0x~16r, gens ~p',
+            [Quorum, PrepNodes, Rps]),
       majority(PrepNodes, Quorum),
-      debug(paxos, 'Prepare: ~p', [Rps]),
       max_list(Rps, K),
       succ(K, K1),
       paxos_message(accept(Key,Na,K1,Ra,Value), TMO, Accept),
@@ -655,14 +767,13 @@ intersecting(Set1, Set2) :-
 %   and NodeSet is the set (bitmask) of   Node values that replies, i.e.
 %   |NodeSet| is length(Result). The transfer stops   if  all members of
 %   the set Quorum responded or the configured timeout passed.
-%
-%   @tbd If we get a `nack` we can stop
 
 collect(Quorum, Stop, Node, Template, Message, Result, NodeSet) :-
     State = state(0),
     L0 = [dummy|_],
     Answers = list(L0),
     (   broadcast_request(Message),
+        debug(paxos(request), 'broadcast_request: ~p', [Message]),
         (   Stop
         ->  !,
             fail
@@ -684,6 +795,34 @@ collect(Quorum, Stop, Node, Template, Message, Result, NodeSet) :-
     arg(1, Answers, [_]),               % close the answer list
     L0 = [_|Result].
 
+%!  paxos_quorum_ask(?Template, +Message, -Result, +Options)
+%
+%   Ask the paxos forum for their opinion.  This predicate is not really
+%   part  of  the  paxos  protocol,  but    reuses  notably  the  quorum
+%   maintenance mechanism of this library for   asking  questions to the
+%   quorum (cluster). Message is the message to   be  asked. Result is a
+%   list of copies of Template from the quorum. Options:
+%
+%     - timeout(+Seconds)
+%       Max time to wait for a reply. Default is the setting
+%       `response_timeout`.
+%     - node(?Node)
+%       Can be used to include the replying node into Template.
+%     - quorum(?Quorum)
+%       Set/query the interviewed quorum as a bitmask
+
+paxos_quorum_ask(Template, Message, Result, Options) :-
+    option(timeout(TMO), Options, TMO),
+    option(node(Node), Options, _),
+    option(quorum(Quorum), Options, Quorum),
+    apply_default(TMO, response_timeout),
+    (   var(Quorum)
+    ->  life_quorum(Quorum, _Alive)
+    ;   true
+    ),
+    paxos_message(ask(Node, Message), TMO, BroadcastMessage),
+    collect(Quorum, false, Node, Template, BroadcastMessage, Result, _PrepNodes).
+
 %!  paxos_get(?Term) is semidet.
 %
 %   Equivalent to paxos_key(Term,Key), pasox_get(Key,Term).   I.e., Term
@@ -808,11 +947,15 @@ replicator :-
     debug(paxos(replicate), 'Starting replicator', []),
     State = state(idle),
     repeat,
-      life_quorum(Quorum),
-      (   Quorum /\ \(1<<Node) =:= 0
-      ->  debug(paxos(replicate), 'I''m alone, waiting ...', []),
+      quorum(Quorum),
+      dead(Dead),
+      LifeQuorum is Quorum /\ \Dead,
+      (   LifeQuorum /\ \(1<<Node) =:= 0
+      ->  debug(paxos(replicate),
+                'Me: ~d, Quorum: 0x~16r, Dead: 0x~16r: I''m alone, waiting ...',
+                [Node, Quorum, Dead]),
           thread_get_message(_)
-      ;   (   paxos_replicate_key(Quorum, Key, [])
+      ;   (   paxos_replicate_key(LifeQuorum, Key, [])
           ->  replicated(State, key(Key)),
               thread_self(Me),
               thread_get_message(Me, _, [timeout(ReplSleep)])
@@ -822,11 +965,6 @@ replicator :-
       ),
       fail.
 
-life_quorum(LifeQuorum) :-
-    quorum(Quorum),
-    dead(Dead),
-    LifeQuorum is Quorum /\ \Dead.
-
 replicated(State, key(_Key)) :-
     arg(1, State, idle),
     !,
@@ -879,7 +1017,8 @@ replication_key(Nodes, Key) :-
 
 needs_replicate(Nodes, Key) :-
     ledger_current(Key, _Gen, _Value, Holders),
-    Nodes /\ \Holders =\= 0.
+    Nodes /\ \Holders =\= 0,
+    \+ admin_key(_, Key).
 
 
 		 /*******************************
@@ -911,11 +1050,21 @@ paxos_on_change(Key, Value, Goal) :-
     must_be(callable, Plain),
     (   Plain == ignore
     ->  unlisten(paxos_user, paxos_changed(Key,Value))
-    ;   listen(paxos_user, paxos(changed(Key,Value)),
-               thread_create(Goal, _, [detached(true)])),
+    ;   listen(paxos_user, paxos_changed(Key,Value),
+               key_changed(Key, Value, Goal)),
         paxos_initialize
     ).
 
+key_changed(_Key, _Value, Goal) :-
+    E = error(_,_),
+    catch(thread_create(Goal, _, [detached(true)]),
+          E, key_error(E)).
+
+key_error(error(permission_error(create, thread, _), _)) :-
+    !.
+key_error(E) :-
+    print_message(error, E).
+
 
 		 /*******************************
 		 *            HOOKS		*
@@ -937,18 +1086,43 @@ paxos_on_change(Key, Value, Goal) :-
 %
 %   @arg TimeOut is one of `-` or a time in seconds.
 
+paxos_message(Paxos:From, TMO, Message) :-
+    paxos_message_raw(paxos(Paxos):From, TMO, Message).
 paxos_message(Paxos, TMO, Message) :-
-    paxos_message_hook(paxos(Paxos), TMO, Message),
+    paxos_message_raw(paxos(Paxos), TMO, Message).
+
+paxos_message_raw(Message, TMO, WireMessage) :-
+    paxos_message_hook(Message, TMO, WireMessage),
     !.
-paxos_message(Paxos, TMO, Message) :-
+paxos_message_raw(Message, TMO, WireMessage) :-
     throw(error(mode_error(det, fail,
-                           paxos:paxos_message_hook(Paxos, TMO, Message)), _)).
+                           paxos:paxos_message_hook(Message, TMO, WireMessage)), _)).
 
 
 		 /*******************************
 		 *           STORAGE		*
 		 *******************************/
 
+%!  paxos_ledger_hook(+Action, ?Key, ?Gen, ?Value, ?Holders)
+%
+%   Hook called for all operations on the ledger.  Defined actions are:
+%
+%     - current
+%       Enumerate our ledger content.
+%     - get
+%       Get a single value from our ledger.
+%     - create
+%       Create a new key in our ledger.
+%     - accept
+%       Accept a new newly proposed value for a key.  Failure causes
+%       the library to send a _NACK_ message.
+%     - set
+%       Final acceptance of Ken@Gen, providing the holders that accepted
+%       the new value.
+%     - learn
+%       Accept new keys in a new node or node that has been offline for
+%       some time.
+
 :- dynamic
     paxons_ledger/4.                    % Key, Gen, Value, Holders
 
@@ -956,6 +1130,8 @@ paxos_message(Paxos, TMO, Message) :-
 %
 %   True when Key is a known key in my ledger.
 
+ledger_current(Key, Gen, Value, Holders) :-
+    paxos_ledger_hook(current, Key, Gen, Value, Holders).
 ledger_current(Key, Gen, Value, Holders) :-
     paxons_ledger(Key, Gen, Value, Holders),
     valid(Holders).
@@ -967,6 +1143,11 @@ ledger_current(Key, Gen, Value, Holders) :-
 %   Note that if the value is  not   yet  acknowledged  by the leader we
 %   should not use it.
 
+ledger(Key, Gen, Value) :-
+    paxos_ledger_hook(get, Key, Gen, Value0, Holders),
+    !,
+    valid(Holders),
+    Value = Value0.
 ledger(Key, Gen, Value) :-
     paxons_ledger(Key, Gen, Value0, Holders),
     valid(Holders),
@@ -978,6 +1159,9 @@ ledger(Key, Gen, Value) :-
 %   Create a new Key-Value pair  at   generation  Gen.  This is executed
 %   during the preparation phase.
 
+ledger_create(Key, Gen, Value) :-
+    paxos_ledger_hook(create, Key, Gen, Value, -),
+    !.
 ledger_create(Key, Gen, Value) :-
     get_time(Now),
     asserta(paxons_ledger(Key, Gen, Value, created(Now))).
@@ -987,6 +1171,9 @@ ledger_create(Key, Gen, Value) :-
 %   Update Key to Value if the  current   generation  is older than Gen.
 %   This reflects the accept phase of the protocol.
 
+ledger_update(Key, Gen, Value) :-
+    paxos_ledger_hook(accept, Key, Gen, Value, -),
+    !.
 ledger_update(Key, Gen, Value) :-
     paxons_ledger(Key, Gen0, _Value, _Holders),
     !,
@@ -1002,6 +1189,9 @@ ledger_update(Key, Gen, Value) :-
 %
 %   The leader acknowledged that Key@Gen represents a valid new
 
+ledger_update_holders(Key, Gen, Holders) :-
+    paxos_ledger_hook(set, Key, Gen, _, Holders),
+    !.
 ledger_update_holders(Key, Gen, Holders) :-
     clause(paxons_ledger(Key, Gen, Value, Holders0), true, Ref),
     !,
@@ -1028,6 +1218,9 @@ clean_key(_, Key, Gen) :-
 %
 %   We received a learn event.
 
+ledger_learn(Key,Gen,Value) :-
+    paxos_ledger_hook(learn, Key, Gen, Value, -),
+    !.
 ledger_learn(Key,Gen,Value) :-
     paxons_ledger(Key, Gen0, Value0, _Holders),
     !,
@@ -1048,9 +1241,11 @@ ledger_learn(Key,Gen,Value) :-
 %   thread.
 
 ledger_forget(Nodes) :-
-    thread_create(ledger_forget_threaded(Nodes), _,
-                  [ detached(true)
-                  ]).
+    catch(thread_create(ledger_forget_threaded(Nodes), _,
+                        [ detached(true)
+                        ]),
+          error(permission_error(create, thread, _), _),
+          true).
 
 ledger_forget_threaded(Nodes) :-
     debug(paxos(node), 'Forgetting 0x~16r', [Nodes]),
