@@ -179,13 +179,15 @@ destroyDefinition(Definition def)
   { DEBUG(MSG_PROC_COUNT, Sdprintf("Unalloc foreign/thread-local: %s\n",
 				   predicateName(def)));
     if ( true(def, P_DIRTYREG) )
-    { DEBUG(0, Sdprintf("Dirty: %s\n", predicateName(def)));
+    { DEBUG(0, if ( GD->cleaning == CLN_NORMAL )
+	         Sdprintf("Destroying dirty predicate: %s\n",
+			  predicateName(def)));
       unregisterDirtyDefinition(def);
     }
 
 #ifdef O_PLMT
     if ( true(def, P_THREAD_LOCAL) )
-      free_ldef_vector(def->impl.local);
+      destroyLocalDefinitions(def);
 #endif
 
     freeHeap(def, sizeof(*def));
@@ -432,6 +434,7 @@ void
 shareDefinition(Definition def)
 { int shared = ATOMIC_INC(&def->shared);
   assert(shared > 0);
+  (void)shared;
 }
 
 
@@ -1377,6 +1380,8 @@ retractClauseDefinition(Definition def, Clause clause)
   deleteActiveClauseFromIndexes(def, clause); /* just updates "dirtyness" */
   def->impl.clauses.number_of_clauses--;
   def->impl.clauses.erased_clauses++;
+  if ( false(clause, UNIT_CLAUSE) )
+    def->impl.clauses.number_of_rules--;
 #ifdef O_LOGICAL_UPDATE
   clause->generation.erased = next_global_generation();
   setLastModifiedPredicate(def, clause->generation.erased);
@@ -1462,16 +1467,9 @@ announceErasedClause(Clause clause)
 
 static void
 freeLingeringDefinition(Definition def, DirtyDefInfo ddi)
-{ linger_list *c;
-
-  do
-  { if ( !(c=def->lingering) )
-      return;
-    if ( ddi->oldest_generation != GEN_MAX )
-      return;
-  } while( !COMPARE_AND_SWAP(&def->lingering, c, NULL) );
-
-  free_lingering(c);
+{ gen_t gen = ddi->oldest_generation == GEN_MAX ? global_generation()
+						: ddi->oldest_generation;
+  free_lingering(&def->lingering, gen);
 }
 
 
